/*
 * The Spar Library - modular math parser
 * Copyright (C) 2000,2001 Davide Angelocola <davide178@inwind.it>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 *
 */

#include <spar/sl_formula.h>
#include <spar/sl_math_library.h>
#include <spar/sl_math.h>

inline double
sl_formula_tan (const double alpha)
{				/* tan(x) = sin(x) / cos(x) */
  return (sl_sin (alpha) / sl_cos (alpha));
}

inline double
sl_formula_cotan (const double alpha)
{				/* cotan(x) = cos(x) / sin(x) */
  return (sl_cos (alpha) / sl_sin (alpha));
}

inline double
sl_formula_sin_add (const double alpha, const double beta)
{				/* sin(x+y) */
  return (sl_sin (alpha) * sl_cos (beta) + sl_cos (alpha) * sl_sin (beta));
}

inline double
sl_formula_sin_sub (const double alpha, const double beta)
{				/* sin(x-y) */
  return (sl_sin (alpha) * sl_cos (beta) - sl_cos (alpha) * sl_sin (beta));
}

inline double
sl_formula_cos_add (const double alpha, const double beta)
{				/* cos(x+y) */
  return (sl_cos (alpha) * sl_cos (beta) - sl_sin (alpha) * sl_sin (beta));
}

inline double
sl_formula_cos_sub (const double alpha, const double beta)
{				/* cos(x-y) */
  return (sl_cos (alpha) * sl_cos (beta) + sl_sin (alpha) * sl_sin (beta));
}

inline double
sl_formula_tan_add (const double alpha, const double beta)
{				/* tan(x+y) */
  double n = sl_tan (alpha) + sl_tan (beta);
  double d = 1. - (sl_tan (alpha) * sl_tan (beta));

  return (n / d);
}

inline double
sl_formula_tan_sub (const double alpha, const double beta)
{				/* tan(x-y) */
  double n = sl_tan (alpha) - sl_tan (beta);
  double d = 1. + (sl_tan (alpha) * sl_tan (beta));

  return (n / d);
}

inline double
sl_formula_sin_dup (const double alpha)
{				/* sin(2x) */
  return (2. * sl_sin (alpha) * sl_sin (alpha));
}

inline double
sl_formula_cos_dup (const double alpha)
{				/* cos(2x) */
  return (sl_pow_2 (sl_cos (alpha)) - sl_pow_2 (sl_sin (alpha)));
}

inline double
sl_formula_tan_dup (const double alpha)
{				/* tan(2x) */
  double n = 2. * sl_tan (alpha);
  double d = 1. - sl_pow_2 (sl_tan (alpha));

  return (n / d);
}

inline double
sl_formula_sin_mul_sin (const double alpha, const double beta)
{				/* sin(a) * sin(b) */
  return 0.5 * (sl_cos (alpha - beta) - sl_cos (alpha + beta));
}

inline double
sl_formula_sin_mul_cos (const double alpha, const double beta)
{				/* sin(a) * cos(b) */
  return 0.5 * (sl_sin (alpha + beta) + sl_sin (alpha - beta));
}

inline double
sl_formula_cos_mul_cos (const double alpha, const double beta)
{				/* cos(a) * cos(b) */
  return 0.5 * (sl_cos (alpha + beta) + sl_cos (alpha - beta));
}

#define MP(a,b) ((a)+(b)/2.)
#define MN(a,b) ((a)-(b)/2.)

inline double
sl_formula_sin_plus_sin (const double alpha, const double beta)
{				/* sin(a) + sin(b) */
  return 2. * (sl_sin (MP (alpha, beta)) * sl_cos (MN (alpha, beta)));
}

inline double
sl_formula_cos_plus_cos (const double alpha, const double beta)
{				/* cos(a) + cos(b) */
  return 2. * (sl_cos (MP (alpha, beta)) * sl_cos (MN (alpha, beta)));
}

inline double
sl_formula_sin_minus_sin (const double alpha, const double beta)
{				/* sin(a) - sin(b) */
  return 2. * (sl_cos (MP (alpha, beta)) * sl_sin (MN (alpha, beta)));
}

inline double
sl_formula_cos_minus_cos (const double alpha, const double beta)
{				/* cos(a) - cos(b) */
  return -2. * (sl_sin (MP (alpha, beta)) * sl_sin (MN (alpha, beta)));
}

inline double
sl_formula_sin_bisec (const double alpha)
{				/* sin(x/2) */
  return sl_sqrt ((1. - sl_cos (alpha)) / 2.);
}

inline double
sl_formula_cos_bisec (const double alpha)
{				/* cos(x/2) */
  return sl_sqrt ((1. + sl_cos (alpha)) / 2.);
}

inline double
sl_formula_tan_bisec (const double alpha)
{				/* tan(x/2) */
  return sl_sqrt ((1. - sl_cos (alpha)) / (1. + sl_cos (alpha)));
}
