
/*
 * The Spar Library - a maths applications framework
 * Copyright (C) 2000,2001 Davide Angelocola <davide178@inwind.it>
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307, USA.
 */

#include <spar/sl_poly.h>
#include <spar/sl_complex.h>
#include <spar/sl_math.h>
#include <spar/sl_math_library.h>

#define SWAP(a,b) do { double tmp = b ; b = a ; a = tmp ; } while(0)

int
sl_poly_zsolve_cubic (double a, double b, double c, 
                              sl_complex *z0, sl_complex *z1, 
                              sl_complex *z2)
{
  double q = (a * a - 3 * b);
  double r = (2 * a * a * a - 9 * a * b + 27 * c);

  double Q = q / 9;
  double R = r / 54;

  double Q3 = Q * Q * Q;
  double R2 = R * R;

  double CR2 = 729 * r * r;
  double CQ3 = 2916 * q * q * q;

  if (R == 0 && Q == 0)
    {
      SL_COMPLEX_RE (*z0) = -a / 3;
      SL_COMPLEX_IM (*z0) = 0;
      SL_COMPLEX_RE (*z1) = -a / 3;
      SL_COMPLEX_IM (*z1) = 0;
      SL_COMPLEX_RE (*z2) = -a / 3;
      SL_COMPLEX_IM (*z2) = 0;
      return 3;
    }
  else if (CR2 == CQ3) 
    {
      double sqrtQ = sl_sqrt (Q);

      if (R > 0)
	{
	  SL_COMPLEX_RE (*z0) = -2 * sqrtQ - a / 3;
	  SL_COMPLEX_IM (*z0) = 0;
	  SL_COMPLEX_RE (*z1) = sqrtQ - a / 3;
	  SL_COMPLEX_IM (*z1) = 0;
	  SL_COMPLEX_RE (*z2) = sqrtQ - a / 3;
	  SL_COMPLEX_IM (*z2) = 0;
	}
      else
	{
	  SL_COMPLEX_RE (*z0) = -sqrtQ - a / 3;
	  SL_COMPLEX_IM (*z0) = 0;
	  SL_COMPLEX_RE (*z1) = -sqrtQ - a / 3;
	  SL_COMPLEX_IM (*z1) = 0;
	  SL_COMPLEX_RE (*z2) = 2 * sqrtQ - a / 3;
	  SL_COMPLEX_IM (*z2) = 0;
	}
      return 3;
    }
  else if (CR2 < CQ3)  
  {
      double sqrtQ = sl_sqrt (Q);
      double sqrtQ3 = sl_pow_3(sqrtQ);
      double theta = sl_arccos (R / sqrtQ3);
      double norm = -2 * sqrtQ;
      double r0 = norm * sl_cos (theta / 3) - a / 3;
      double r1 = norm * sl_cos ((theta + 2.0 * M_PI) / 3) - a / 3;
      double r2 = norm * sl_cos ((theta - 2.0 * M_PI) / 3) - a / 3;

      if (r0 > r1)
	SWAP (r0, r1);

      if (r1 > r2)
	{
	  SWAP (r1, r2);

	  if (r0 > r1)
	    SWAP (r0, r1);
	}

      SL_COMPLEX_RE (*z0) = r0;
      SL_COMPLEX_IM (*z0) = 0;

      SL_COMPLEX_RE (*z1) = r1;
      SL_COMPLEX_IM (*z1) = 0;

      SL_COMPLEX_RE (*z2) = r2;
      SL_COMPLEX_IM (*z2) = 0;

      return 3;
    }
  else
    {
      double sgnR = (R >= 0 ? 1 : -1);
      double A = -sgnR * sl_pow (sl_abs (R) + sl_sqrt (R2 - Q3), 1.0 / 3.0);
      double B = Q / A;

      if (A + B < 0)
	{
	  SL_COMPLEX_RE (*z0) = A + B - a / 3;
	  SL_COMPLEX_IM (*z0) = 0;

	  SL_COMPLEX_RE (*z1) = -0.5 * (A + B) - a / 3;
	  SL_COMPLEX_IM (*z1) = -(sl_sqrt (3.0) / 2.0) * sl_abs(A - B);

	  SL_COMPLEX_RE (*z2) = -0.5 * (A + B) - a / 3;
	  SL_COMPLEX_IM (*z2) = (sl_sqrt (3.0) / 2.0) * sl_abs(A - B);
	}
      else
	{
	  SL_COMPLEX_RE (*z0) = -0.5 * (A + B) - a / 3;
	  SL_COMPLEX_IM (*z0) = -(sl_sqrt (3.0) / 2.0) * sl_abs(A - B);

	  SL_COMPLEX_RE (*z1) = -0.5 * (A + B) - a / 3;
	  SL_COMPLEX_IM (*z1) = (sl_sqrt (3.0) / 2.0) * sl_abs(A - B);

	  SL_COMPLEX_RE (*z2) = A + B - a / 3;
	  SL_COMPLEX_IM (*z2) = 0;
	}

      return 3;
    }
}
